module Tiramisu
  class Unit

    # @example
    #   assert(x) == y
    #   assert(x).include?(y)
    #   assert(x).nil?
    #   refute(a).include?(b)
    #   fail_if(a).include?(b)
    [
      :assert,
      :refute
    ].each do |meth|
      define_method meth do |obj = nil, &block|
        __objects__[obj = block ? Tiramisu.call_block(block) : {returned: obj}.freeze] = true
        __assertions__.push(Tiramisu::Assert.new(obj, meth, caller[0])).last
      end
    end
    alias expect assert
    alias fail_if refute

    # stop executing current test and mark it as skipped
    #
    # @example
    #   test :something do
    #     skip "recheck this after fixing X"
    #     assert(x) == y # this wont run
    #   end
    #
    def skip reason = nil
      throw(:__tiramisu_status__, Tiramisu::Skip.new(reason, caller[0]))
    end

    def __run__ test, before, around, after
      __send__(before) if before
      if around
        __send__(around, proc {__send__(test)})
      else
        __send__(test)
      end
      __send__(after) if after
      __assertions__.each(&:__validate_expectations__)
      __objects__.each_key {|o| Tiramisu::Expectation.restore_object_status(o[:returned])}
      :__tiramisu_passed__
    rescue Exception => e
      throw(:__tiramisu_status__, e)
    end

    private
    def __objects__
      @__objects__ ||= {}
    end

    def __assertions__
      @__assertions__ ||= []
    end
  end

  class << Unit

    def inherited base
      base.instance_variable_set(:@__tiramisu_hooks__, Marshal.load(Marshal.dump(hooks)))
    end

    def spec
      raise(NotImplementedError, 'Nested specs not supported. Please use a context instead')
    end

    # define a context inside current spec/context.
    #
    # @param label
    # @param &block
    #
    def context label, &block
      return unless block
      Tiramisu.define_and_register_a_context(label, block, self)
    end

    # define a test
    #
    # @param label
    # @param &block
    #
    def test label, &block
      return unless block
      tests[label] = Tiramisu.identity_string(:test, label, block)
      define_method(tests[label], &block)
    end
    alias it test
    alias should test

    def tests
      @__tiramisu_tests__ ||= {}
    end

    # run some code before/around/after tests
    #
    # @example
    #   spec User do
    #     before do
    #       @users = ...
    #     end
    #
    #     test :login do
    #       # @users available here
    #     end
    #   end
    #
    # @example
    #   spec User do
    #     around |&test|
    #       SomeApi.with_auth do
    #         test.call # run the test
    #       end
    #     end
    #
    #     test :authorize do
    #       # will run inside `with_auth` block
    #     end
    #   end
    #
    # @note hooks are inherited from parent spec or context
    #
    # @example
    #   spec Math do
    #     before { @n = rand }
    #
    #     context :PI do
    #       test :some_test do
    #         # @n available here
    #       end
    #     end
    #   end
    #
    # @note defined hooks will override inherited ones
    #
    # @example
    #   spec Math do
    #     before { @n = 1 }
    #
    #     context :x do
    #       before { @n = 0 }
    #       test :some_test do
    #         # @n is zero here
    #       end
    #     end
    #   end
    #
    [
      :before,
      :around,
      :after
    ].each do |hook|
      define_method hook do |&block|
        block || raise(ArgumentError, 'block missing')
        hooks[hook] = :"__tiramisu__#{hook}_hook__"
        define_method(hooks[hook], &block)
      end
    end

    def hooks
      @__tiramisu_hooks__ ||= {}
    end

    # skipping a whole spec/context
    #
    # @example
    #   spec Array do
    #     skip
    #
    #     # code here wont be executed
    #   end
    #
    def skip reason = nil
      throw(:__tiramisu_skip__, Tiramisu::Skip.new(reason, caller[0]))
    end

    def run test
      tests[test] || raise(StandardError, 'Undefined test %s at "%s"' % [test.inspect, __identity__])
      catch :__tiramisu_status__ do
        allocate.__run__(tests[test], hooks[:before], hooks[:around], hooks[:after])
      end
    end
  end
end
